/*****************************************************************************
 * interface.c: interface access for other threads
 * This library provides basic functions for threads to interact with user
 * interface, such as command line.
 *****************************************************************************
 * Copyright (C) 1998-2001 VideoLAN
 * $Id: interface.c,v 1.93 2002/04/23 14:16:21 sam Exp $
 *
 * Authors: Vincent Seguin <seguin@via.ecp.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include <errno.h>                                                 /* ENOMEM */
#include <stdlib.h>                                      /* free(), strtol() */
#include <stdio.h>                                                   /* FILE */
#include <string.h>                                            /* strerror() */
#include <sys/types.h>                                              /* off_t */

#include <videolan/vlc.h>

#include "stream_control.h"
#include "input_ext-intf.h"

#include "audio_output.h"

#include "interface.h"
#include "intf_playlist.h"

#include "video.h"
#include "video_output.h"

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static void intf_Manage( intf_thread_t *p_intf );

/*****************************************************************************
 * intf_Create: prepare interface before main loop
 *****************************************************************************
 * This function opens output devices and creates specific interfaces. It sends
 * its own error messages.
 *****************************************************************************/
intf_thread_t* intf_Create( vlc_object_t *p_parent )
{
    intf_thread_t * p_intf;
    char *psz_name;

    /* Allocate structure */
    p_intf = malloc( sizeof( intf_thread_t ) );
    if( !p_intf )
    {
        intf_ErrMsg( "intf error: cannot create interface thread (%s)",
                     strerror( ENOMEM ) );
        return( NULL );
    }

    VLC_INIT_OBJECT( p_intf, p_parent );

    /* Choose the best module */
    psz_name = config_GetPszVariable( p_intf->p_this, "intf" );
    p_intf->p_module = module_Need( p_intf->p_this, MODULE_CAPABILITY_INTF,
                                    psz_name, (void *)p_intf );

    if( psz_name ) free( psz_name );
    if( p_intf->p_module == NULL )
    {
        intf_ErrMsg( "intf error: no suitable intf module" );
        free( p_intf );
        return( NULL );
    }

#define f p_intf->p_module->p_functions->intf.functions.intf
    p_intf->pf_open       = f.pf_open;
    p_intf->pf_close      = f.pf_close;
    p_intf->pf_run        = f.pf_run;
#undef f

    /* Initialize callbacks */
    p_intf->pf_manage     = intf_Manage;

    /* Initialize structure */
    p_intf->b_menu        = 0;
    p_intf->b_menu_change = 0;

    /* Initialize mutexes */
    vlc_mutex_init( p_intf->p_this, &p_intf->change_lock );

    intf_WarnMsg( 1, "intf: interface initialized");
    return( p_intf );
}

/*****************************************************************************
 * intf_RunThread: launch the interface thread
 *****************************************************************************
 * This function creates a new thread and runs the interface in it.
 *****************************************************************************/
vlc_thread_t intf_RunThread( intf_thread_t *p_intf )
{
    vlc_thread_t thread_id;

    if( vlc_thread_create( p_intf->p_this, &thread_id, "interface",
         (vlc_thread_func_t)p_intf->pf_run, (void *)p_intf ) )
    {
        intf_ErrMsg( "intf error: can't spawn interface thread" );
        return 0;
    }

    return thread_id;
}

/*****************************************************************************
 * intf_StopThread: end the interface thread
 *****************************************************************************
 * This function asks the interface thread to stop.
 *****************************************************************************/
void intf_StopThread( intf_thread_t *p_intf )
{
    /* Tell the interface to die */
    p_intf->p_vlc->b_die = 1;

    /* Wait for the thread to exit */
    vlc_thread_join( p_intf->thread_id );
}

/*****************************************************************************
 * intf_Destroy: clean interface after main loop
 *****************************************************************************
 * This function destroys specific interfaces and close output devices.
 *****************************************************************************/
void intf_Destroy( intf_thread_t *p_intf )
{
    /* Destroy interface */
    p_intf->pf_close( p_intf );

#if 0
    /* Close input thread, if any (blocking) */
    if( p_intf->p_input )
    {   
        input_DestroyThread( p_intf->p_input, NULL );
    }
#endif

    /* Unlock module */
    module_Unneed( p_intf->p_module );

    vlc_mutex_destroy( &p_intf->change_lock );

    /* Free structure */
    free( p_intf );
}

/*****************************************************************************
 * intf_Manage: manage interface
 *****************************************************************************
 * This function has to be called regularly by the interface plugin. It
 * checks for playlist end, module expiration, message flushing, and a few
 * other useful things.
 *****************************************************************************/
static void intf_Manage( intf_thread_t *p_intf )
{
    /* Manage module bank */
    module_ManageBank( p_intf->p_this );

    vlc_mutex_lock( &p_intf->p_vlc->p_input_bank->lock );

    if( p_intf->p_vlc->p_input_bank->i_count )
    {
        int i_input;
        input_thread_t *p_input;

        for( i_input = 0; i_input < p_intf->p_vlc->p_input_bank->i_count; i_input++ )
        {
            p_input = p_intf->p_vlc->p_input_bank->pp_input[i_input];
            
            if( p_input->i_status == THREAD_OVER )
            {
                input_DestroyThread( p_input );
                p_intf->p_vlc->p_input_bank->pp_input[i_input] = NULL;
                p_intf->p_vlc->p_input_bank->i_count--;
            }
            else if( ( p_input->i_status == THREAD_READY
                        || p_input->i_status == THREAD_ERROR )
                     && ( p_input->b_error || p_input->b_eof ) )
            {
                input_StopThread( p_input, NULL );
            }
        }

        vlc_mutex_unlock( &p_intf->p_vlc->p_input_bank->lock );
    }
    /* If no stream is being played, try to find one */
    else
    {
//        vlc_mutex_lock( &p_intf->p_vlc->p_playlist->change_lock );

        if( !p_intf->p_vlc->p_playlist->b_stopped )
        {
            /* Select the next playlist item */
            intf_PlaylistNext( p_intf->p_vlc->p_playlist );

            /* don't loop by default: stop at playlist end */
            if( p_intf->p_vlc->p_playlist->i_index == -1 )
            {
                p_intf->p_vlc->p_playlist->b_stopped = 1;
            }
            else
            {
                input_thread_t *p_input;

                p_intf->p_vlc->p_playlist->b_stopped = 0;
                p_intf->p_vlc->p_playlist->i_mode = PLAYLIST_FORWARD + 
                    config_GetIntVariable( p_intf->p_this, "loop-playlist" );
                intf_WarnMsg( 3, "intf: creating new input thread" );
                p_input = input_CreateThread( p_intf->p_this,
                              &p_intf->p_vlc->p_playlist->current, NULL );
                if( p_input != NULL )
                {
                    p_intf->p_vlc->p_input_bank->pp_input[ p_intf->p_vlc->p_input_bank->i_count ] = p_input;
                    p_intf->p_vlc->p_input_bank->i_count++;
                }
            }

            vlc_mutex_unlock( &p_intf->p_vlc->p_input_bank->lock );
        }
        else
        {
            vlc_mutex_unlock( &p_intf->p_vlc->p_input_bank->lock );

            /* playing has been stopped: we no longer need outputs */
            if( p_intf->p_vlc->p_aout_bank->i_count )
            {
                /* FIXME kludge that does not work with several outputs */
                aout_DestroyThread( p_intf->p_vlc->p_aout_bank->pp_aout[0], NULL );
                p_intf->p_vlc->p_aout_bank->i_count--;
            }
            if( p_intf->p_vlc->p_vout_bank->i_count )
            {
                vout_DestroyThread( p_intf->p_vlc->p_vout_bank->pp_vout[0], NULL );
                p_intf->p_vlc->p_vout_bank->i_count--;
            }
        }

//        vlc_mutex_unlock( &p_intf->p_vlc->p_playlist->change_lock );
    }
}

